package ru.dbotthepony.kstarbound.util

import com.google.common.util.concurrent.Futures
import java.util.concurrent.Callable
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.ExecutorService
import java.util.concurrent.Future
import java.util.concurrent.FutureTask
import java.util.concurrent.TimeUnit
import java.util.concurrent.locks.LockSupport

class ManualExecutorService(val thread: Thread = Thread.currentThread()) : ExecutorService {
	private val executeQueue = ConcurrentLinkedQueue<Runnable>()
	private val futureQueue = ConcurrentLinkedQueue<FutureTask<*>>()

	fun isSameThread(): Boolean {
		return Thread.currentThread() === thread
	}

	fun executeQueuedTasks() {
		check(isSameThread()) { "Trying to execute queued tasks in thread ${Thread.currentThread()}, while correct thread is $thread" }

		var next = executeQueue.poll()

		while (next != null) {
			next.run()
			next = executeQueue.poll()
		}

		var next2 = futureQueue.poll()

		while (next2 != null) {
			next2.run()
			Thread.interrupted()
			next2 = futureQueue.poll()
		}
	}

	override fun execute(command: Runnable) {
		if (isSameThread()) {
			command.run()
		} else {
			executeQueue.add(command)
			LockSupport.unpark(thread)
		}
	}

	override fun shutdown() {
		throw UnsupportedOperationException()
	}

	override fun shutdownNow(): MutableList<Runnable> {
		throw UnsupportedOperationException()
	}

	override fun isShutdown(): Boolean {
		return false
	}

	override fun isTerminated(): Boolean {
		return false
	}

	override fun awaitTermination(timeout: Long, unit: TimeUnit): Boolean {
		throw UnsupportedOperationException()
	}

	override fun <T : Any?> submit(task: Callable<T>): Future<T> {
		if (isSameThread()) return Futures.immediateFuture(task.call())
		return FutureTask(task).also { futureQueue.add(it); LockSupport.unpark(thread) }
	}

	override fun <T : Any?> submit(task: Runnable, result: T): Future<T> {
		if (isSameThread()) { task.run(); return Futures.immediateFuture(result) }
		return FutureTask { task.run(); result }.also { futureQueue.add(it); LockSupport.unpark(thread) }
	}

	override fun submit(task: Runnable): Future<*> {
		if (isSameThread()) { task.run(); return Futures.immediateVoidFuture() }
		return FutureTask { task.run() }.also { futureQueue.add(it); LockSupport.unpark(thread) }
	}

	override fun <T : Any?> invokeAll(tasks: Collection<Callable<T>>): List<Future<T>> {
		if (isSameThread()) {
			return tasks.map { Futures.immediateFuture(it.call()) }
		} else {
			return tasks.map { submit(it) }.onEach { it.get() }
		}
	}

	override fun <T : Any?> invokeAll(
		tasks: Collection<Callable<T>>,
		timeout: Long,
		unit: TimeUnit
	): List<Future<T>> {
		if (isSameThread()) {
			return tasks.map { Futures.immediateFuture(it.call()) }
		} else {
			return tasks.map { submit(it) }.onEach { it.get(timeout, unit) }
		}
	}

	override fun <T : Any?> invokeAny(tasks: Collection<Callable<T>>): T {
		if (tasks.isEmpty())
			throw NoSuchElementException("Provided task list is empty")

		if (isSameThread()) {
			return tasks.first().call()
		} else {
			return submit(tasks.first()).get()
		}
	}

	override fun <T : Any?> invokeAny(tasks: Collection<Callable<T>>, timeout: Long, unit: TimeUnit): T {
		if (tasks.isEmpty())
			throw NoSuchElementException("Provided task list is empty")

		if (isSameThread()) {
			return tasks.first().call()
		} else {
			return submit(tasks.first()).get(timeout, unit)
		}
	}

	fun <V> join(future: Future<V>): V {
		if (!isSameThread())
			return future.get()

		while (!future.isDone) {
			executeQueuedTasks()
			LockSupport.parkNanos(1_000_000L)
		}

		return future.get()
	}
}
